@@ -34,6 +34,7 @@ struct _PangoContext
 {
   GObject parent_instance;
 
+  PangoLanguage *set_language;
   PangoLanguage *language;
   PangoDirection base_dir;
   PangoGravity base_gravity;
@@ -64,7 +65,8 @@ pango_context_init (PangoContext *context)
   context->resolved_gravity = context->base_gravity = PANGO_GRAVITY_SOUTH;
   context->gravity_hint = PANGO_GRAVITY_HINT_NATURAL;
 
-  context->language = NULL;
+  context->set_language = NULL;
+  context->language = pango_language_get_default ();
   context->font_map = NULL;
 
   context->font_desc = pango_font_description_new ();
@@ -278,7 +280,7 @@ pango_context_list_families (PangoContext          *context,
  * Loads the font in one of the fontmaps in the context
  * that is the closest match for @desc.
  *
- * Returns the font loaded, or %NULL if no font matched.
+ * Returns: the font loaded, or %NULL if no font matched.
  **/
 PangoFont *
 pango_context_load_font (PangoContext               *context,
@@ -299,7 +301,7 @@ pango_context_load_font (PangoContext               *context,
  * Load a set of fonts in the context that can be used to render
  * a font matching @desc.
  *
- * Returns the fontset, or %NULL if no font matched.
+ * Returns: the fontset, or %NULL if no font matched.
  **/
 PangoFontset *
 pango_context_load_fontset (PangoContext               *context,
@@ -363,7 +365,11 @@ pango_context_set_language (PangoContext *context,
 {
   g_return_if_fail (context != NULL);
 
-  context->language = language;
+  context->set_language = language;
+  if (language)
+    context->language = language;
+  else
+    context->language = pango_language_get_default ();
 }
 
 /**
@@ -379,7 +385,7 @@ pango_context_get_language (PangoContext *context)
 {
   g_return_val_if_fail (context != NULL, NULL);
 
-  return context->language;
+  return context->set_language;
 }
 
 /**
@@ -855,7 +861,7 @@ itemize_state_init (ItemizeState      *state,
     state->font_desc_gravity = PANGO_GRAVITY_AUTO;
 
   state->gravity = PANGO_GRAVITY_AUTO;
-  state->centered_baseline = FALSE;
+  state->centered_baseline = PANGO_GRAVITY_IS_VERTICAL (state->context->resolved_gravity);
   state->gravity_hint = state->context->gravity_hint;
   state->resolved_gravity = PANGO_GRAVITY_AUTO;
   state->derived_lang = NULL;
@@ -1013,6 +1019,7 @@ itemize_state_add_character (ItemizeState     *state,
 
   state->item->analysis.flags = state->centered_baseline ? PANGO_ANALYSIS_FLAG_CENTERED_BASELINE : 0;
 
+  state->item->analysis.script = state->script;
   state->item->analysis.language = state->derived_lang;
 
   if (state->copy_extra_attrs)
@@ -1087,6 +1094,21 @@ get_base_font (ItemizeState *state)
   return state->base_font;
 }
 
+static PangoScript
+get_script (ItemizeState      *state)
+{
+  /* Always use a basic shaper for vertical layout (ie, east/west gravity)
+   * as none of our script shapers support vertical shaping right now.
+   *
+   * XXX Should move the knowledge into the shaper interface.
+   */
+
+  if (PANGO_GRAVITY_IS_VERTICAL (state->resolved_gravity))
+    return PANGO_SCRIPT_COMMON;
+  else
+    return state->script;
+}
+
 static gboolean
 get_shaper_and_font (ItemizeState      *state,
 		     gunichar           wc,
@@ -1101,20 +1123,8 @@ get_shaper_and_font (ItemizeState      *state,
     return *shape_engine != NULL;
 
   if (!state->exact_engines && !state->fallback_engines)
-    {
-      /* Always use a basic shaper for vertical layout (ie, east/west gravity)
-       * as we do not support vertical shaping as of now.
-       */
-      PangoScript script;
-
-      if (PANGO_GRAVITY_IS_VERTICAL (state->resolved_gravity))
-	script = PANGO_SCRIPT_COMMON;
-      else
-	script = state->script;
-
-      get_engines (state->context, state->derived_lang, script,
-		   &state->exact_engines, &state->fallback_engines);
-    }
+    get_engines (state->context, state->derived_lang, get_script (state),
+		 &state->exact_engines, &state->fallback_engines);
 
   info.lang = state->derived_lang;
   info.wc = wc;
@@ -1220,7 +1230,6 @@ itemize_state_update_for_new_run (ItemizeState *state)
       if (state->font_desc_gravity != PANGO_GRAVITY_AUTO)
 	{
 	  state->resolved_gravity = state->font_desc_gravity;
-	  state->centered_baseline = PANGO_GRAVITY_IS_VERTICAL (state->resolved_gravity);
 	}
       else
 	{
@@ -1230,7 +1239,6 @@ itemize_state_update_for_new_run (ItemizeState *state)
 	  if (G_LIKELY (gravity == PANGO_GRAVITY_AUTO))
 	    gravity = state->context->resolved_gravity;
 
-	  state->centered_baseline = PANGO_GRAVITY_IS_VERTICAL (gravity);
 	  state->resolved_gravity = pango_gravity_get_for_script (state->script,
 								  gravity,
 								  gravity_hint);
@@ -1284,6 +1292,21 @@ itemize_state_update_for_new_run (ItemizeState *state)
     }
 }
 
+static const char *
+string_from_script (PangoScript script)
+{
+  static GEnumClass *class = NULL;
+  GEnumValue *value;
+  if (!class)
+    class = g_type_class_ref (PANGO_TYPE_SCRIPT);
+
+  value = g_enum_get_value (class, script);
+  if (!value)
+    return string_from_script (PANGO_SCRIPT_INVALID_CODE);
+
+  return value->value_nick;
+}
+
 static void
 itemize_state_process_run (ItemizeState *state)
 {
@@ -1346,8 +1369,30 @@ itemize_state_process_run (ItemizeState *state)
       PangoEngineShape *shape_engine;
       PangoFont *font;
 
-      if (!get_shaper_and_font (state, ' ', &shape_engine, &font))
+      if (G_UNLIKELY (!get_shaper_and_font (state, ' ', &shape_engine, &font)))
 	{
+	  /* If no shaper was found, warn only once per fontmap/script pair */
+	  PangoFontMap *fontmap = state->context->font_map;
+	  const char *script_name = string_from_script (get_script (state));
+
+	  if (!g_object_get_data (G_OBJECT (fontmap), script_name))
+	    {
+	      const char *what;
+	      if (shape_engine == NULL)
+	        what = "shape engine";
+	      else if (font == NULL)
+	        what = "font";
+	      else
+	        what = "nothing (oops!)";
+	       
+	      g_warning ("failed to find shape engine, expect ugly output. engine-type='%s', script='%s'",
+			 pango_font_map_get_shape_engine_type (fontmap),
+			 script_name);
+
+	      g_object_set_data_full (G_OBJECT (fontmap), script_name,
+				      GINT_TO_POINTER (1), NULL);
+	    }
+
 	  shape_engine = _pango_get_fallback_shaper ();
 	  font = NULL;
 	}
@@ -1390,7 +1435,9 @@ itemize_state_finish (ItemizeState *state)
  * computing bidirectional levels (see pango_context_set_base_dir ()),
  * is specified explicitly rather than gotten from the #PangoContext.
  *
- * Return value: a #GList of #PangoItem structures.
+ * Return value: a #GList of #PangoItem structures.  The items should be
+ * freed using pango_item_free() probably in combination with g_list_foreach(),
+ * and the list itself using g_list_free().
  *
  * Since: 1.4
  */
